import os
from math import prod

import numpy as np
import dask.array as da

from glue.core import Data
from glue.config import data_factory
from glue.core.coordinates import AffineCoordinates

from astropy.table import Table

from casa_formats_io.casa_low_level_io.table import CASATable

__all__ = ['read_spectral_cube', 'parse_spectral_cube']


POLARIZATIONS = ['I', 'Q', 'U', 'V',
                 'RR', 'RL', 'LR', 'LL',
                 'XX', 'XY', 'YX', 'YY']


def is_casa_table(filename):
    if (isinstance(filename, str) and
            os.path.isdir(filename) and
            os.path.exists(os.path.join(filename, 'table.dat'))):
        with open(os.path.join(filename, 'table.dat'), 'rb') as f:
            return f.read(4) == b'\xbe\xbe\xbe\xbe'


def table_to_glue_data(table, label):

    # Glue can actually understand astropy tables, but doesn't work well with
    # some columns being vector columns, nor complex numbers, so we expand
    # some of the columns here.

    # Split out vector columns into single columns unless all vector columns
    # are the same

    reference_shape = table[table.colnames[0]].shape
    for colname in table.colnames[1:]:
        if table[colname].shape != reference_shape:
            split = True
            break
    else:
        split = False

    if split:
        for colname in table.colnames:
            if table[colname].ndim > 1:
                if prod(table[colname].shape[1:]) > 100:
                    raise ValueError(f"Table {colname} is too wide to expand into single columns: {table[colname].shape}")
            if table[colname].ndim == 2:
                for i in range(table[colname].shape[1]):
                    table[f'{colname}[{i}]'] = table[colname][:, i]
                table.remove_column(colname)
            elif table[colname].ndim == 3:
                for i in range(table[colname].shape[1]):
                    for j in range(table[colname].shape[2]):
                        table[f'{colname}[{i},{j}]'] = table[colname][:, i, j]
                table.remove_column(colname)

    # Split out complex columns into amp/phase/real/imag
    for colname in table.colnames:
        if table[colname].dtype.kind == 'c':
            table[colname + '.amp'] = np.abs(table[colname])
            table[colname + '.phase'] = da.angle(table[colname])
            table[colname + '.real'] = da.real(table[colname])
            table[colname + '.imag'] = da.imag(table[colname])
            table.remove_column(colname)

    kwargs = dict((c, table[c]) for c in table.colnames)

    data = Data(label=label, **kwargs)

    return data


def ms_table_to_glue_data(table, label, polarizations):

    # For MS tables specifically we can be smarter about how we split columns
    # as we can special case different columns.

    data_shape = table['DATA'].shape

    # Split UVW vector column into U, V, and W
    table['U'] = table['UVW'][:, 0]
    table['V'] = table['UVW'][:, 1]
    table['W'] = table['UVW'][:, 2]
    table.remove_column('UVW')

    # Columns that have polarization dimensions can be split
    for colname in ['FLAG', 'WEIGHT', 'SIGMA', 'DATA', 'MODEL_DATA', 'CORRECTED_DATA']:
        if colname in table.colnames:
            for ipol, pol in enumerate(polarizations):
                table[colname + '_' + pol] = table[colname][..., ipol]
            table.remove_column(colname)

    # Now we are left with some 1-D and some 2-D columns, so we broadcast all
    # the 1-D ones to match the 2-D ones.
    for colname in table.colnames:
        if table[colname].ndim == 1:
            reshaped_data = da.broadcast_to(table[colname], data_shape[:2][::-1]).T
            table.remove_column(colname)
            table[colname] = reshaped_data

    # Split out complex columns into amp/phase/real/imag
    for colname in table.colnames:
        if table[colname].dtype.kind == 'c':
            table[colname + '.amp'] = np.abs(table[colname])
            table[colname + '.phase'] = da.angle(table[colname])
            table[colname + '.real'] = da.real(table[colname])
            table[colname + '.imag'] = da.imag(table[colname])
            table.remove_column(colname)

    kwargs = dict((c, table[c]) for c in table.colnames)

    # For now just set up an identity transform but with correct axis labels
    # to avoid confusion.
    coords = AffineCoordinates(np.identity(3), labels=['Frequency', 'Row index'])

    data = Data(label=label, coords=coords, **kwargs)

    return data

# We set the priority here to make sure that we rank higher than the default
# astropy Table factory.
@data_factory(label='CASA Table', identifier=is_casa_table, priority=1)
def read_casa_table(filename, **kwargs):

    casa_table = CASATable.read(filename)

    datasets = []
    label_prefix = os.path.basename(filename)

    if casa_table.has_data_desc_id():

        # Extract just the DATA_DESC_ID which can be done as a single table if we
        # get just that column
        data_desc_ids = np.sort(np.unique(casa_table.as_astropy_table(include_columns=['DATA_DESC_ID'])['DATA_DESC_ID']))

        # Load in polarization and spectral window tables
        table_desc = Table.read(os.path.join(filename, 'DATA_DESCRIPTION'), format='casa-table')
        table_pol = Table.read(os.path.join(filename, 'POLARIZATION'), format='casa-table')
        # table_spw = Table.read(os.path.join(filename, 'SPECTRAL_WINDOW'), format='casa-table')

        for data_desc_id in data_desc_ids:

            # Extract polarizations
            pol_id = table_desc['POLARIZATION_ID'][data_desc_id]
            corr_type = table_pol['CORR_TYPE'][pol_id]
            polarizations = [POLARIZATIONS[ct - 1] for ct in corr_type]

            # Extract frequencies
            # spw_id = table_desc['SPECTRAL_WINDOW_ID'][data_desc_id]
            # chan_freq = table_spw['CHAN_FREQ'][spw_id]

            table = casa_table.as_astropy_table(data_desc_id=data_desc_id)
            datasets.append(ms_table_to_glue_data(table, label_prefix + f' [DATA_DESC_ID={data_desc_id}]', polarizations))

    else:

        table = casa_table.as_astropy_table()
        datasets.append(table_to_glue_data(table, label_prefix))

    return datasets


def setup():
    # This function doesn't need to do anything - it just needs to be present
    # but when this file is loaded the data factory is automatically registered
    pass



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'casa-formats-io'
copyright = '2020, Thomas Robitaille, Adam Ginsburg, and Eric Koch'
author = 'Thomas Robitaille, Adam Ginsburg, and Eric Koch'

# The full version, including alpha/beta/rc tags
release = '0.0'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['numpydoc', 'sphinx_automodapi.automodapi', 'sphinx.ext.intersphinx']
numpydoc_show_class_members = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'astropy': ('https://docs.astropy.org/en/stable/', None)}

nitpicky = True



import numpy as np
from astropy.wcs import WCS
from astropy.io import fits
from astropy.time import Time

__all__ = ['coordsys_to_astropy_wcs']

EQUATORIAL_SYSTEMS = ['B1950', 'B1950_VLA', 'J2000', 'ICRS']

SPECSYS = {}
SPECSYS['BARY'] = 'BARYCENT'
SPECSYS['TOPO'] = 'TOPOCENT'
SPECSYS['LSRK'] = 'LSRK'
SPECSYS['LSRD'] = 'LSRD'
SPECSYS['GEO'] = 'GEOCENTR'
SPECSYS['GALACTO'] = 'GALACTOC'
SPECSYS['LGROUP'] = 'LOCALGRP'
SPECSYS['CMB'] = 'CMBDIPOL'
SPECSYS['REST'] = 'SOURCE'

RADESYS = {}
RADESYS['J2000'] = 'FK5'
RADESYS['B1950'] = 'FK4'
RADESYS['B1950_VLA'] = 'FK4'
RADESYS['ICRS'] = 'ICRS'

EQUINOX = {}
EQUINOX['J2000'] = 2000.
EQUINOX['B1950'] = 1950.
EQUINOX['B1950_VLA'] = 1979.9

AXES_TO_CTYPE = {
    'Frequency': 'FREQ',
    'GALACTIC': {'Longitude': 'GLON',
                 'Latitude': 'GLAT'},
    'SUPERGAL': {'Longitude': 'SLON',
                 'Latitude': 'SLAT'},
    'ECLIPTIC': {'Longitude': 'ELON',
                 'Latitude': 'ELAT'}
}

for system in EQUATORIAL_SYSTEMS:
    AXES_TO_CTYPE[system] = {'Right Ascension': 'RA--',
                             'Declination': 'DEC-'}


def sanitize_unit(unit):
    if unit == "'":
        return 'arcmin'
    elif unit == '"':
        return 'arcsec'
    else:
        return unit


def coordsys_to_astropy_wcs(coordsys):
    """
    Convert a casac.coordsys object into an :class:`~astropy.wcs.WCS` object
    """

    # Rather than try and parse the CASA coordsys ourselves, we delegate
    # to CASA by getting it to write out a FITS file and reading back in
    # using WCS

    header = fits.Header()

    # Observer information (ObsInfo.cc)

    header['OBSERVER'] = coordsys['observer']
    header['TELESCOP'] = coordsys['telescope']

    if coordsys['obsdate']['refer'] == 'LAST':

        header['TIMESYS'] = 'UTC'

    else:

        header['TIMESYS'] = coordsys['obsdate']['refer']
        header['MJD-OBS'] = coordsys['obsdate']['m0']['value']

        dateobs = Time(header['MJD-OBS'],
                       format='mjd',
                       scale=coordsys['obsdate']['refer'].lower())
        dateobs.precision = 6
        header['DATE-OBS'] = dateobs.isot

    if 'telescopeposition' in coordsys:

        obsgeo_lon = coordsys['telescopeposition']['m0']['value']
        obsgeo_lat = coordsys['telescopeposition']['m1']['value']
        obsgeo_alt = coordsys['telescopeposition']['m2']['value']

        header['OBSGEO-X'] = obsgeo_alt * np.cos(obsgeo_lon) * np.cos(obsgeo_lat)
        header['OBSGEO-Y'] = obsgeo_alt * np.sin(obsgeo_lon) * np.cos(obsgeo_lat)
        header['OBSGEO-Z'] = obsgeo_alt * np.sin(obsgeo_lat)

    # World coordinates

    # Find worldmap entries

    worldmap = {}

    for key, value in coordsys.items():
        if key.startswith('worldmap'):
            index = int(key[8:])
            worldmap[index] = value

    # Now iterate through the different coordinate types to populate the WCS

    header['WCSAXES'] = np.max([np.max(idx) + 1 for idx in worldmap.values()])

    # Initialize PC
    for ii in range(header['WCSAXES']):
        for jj in range(header['WCSAXES']):
            header[f'PC{ii+1}_{jj+1}'] = 0.

    for coord_type in ('direction', 'spectral', 'stokes', 'linear'):

        # Each coordinate type is stored with a numerical index,
        # so for example a cube might have direction0 and
        # spectral1, or spectral0, direction1, stokes2. The actual
        # index is irrelevant for the rest of the loop.
        for index in range(len(worldmap)):
            if f'{coord_type}{index}' in coordsys:
                data = coordsys[f'{coord_type}{index}']
                break
        else:
            # Skip the rest of the loop for the current coord_type
            # since the coord_type wasn't found with any index.
            continue

        if coord_type == 'direction':
            idx1, idx2 = worldmap[index] + 1
            header[f'CTYPE{idx1}'] = AXES_TO_CTYPE[data['system']][data['axes'][0]] + '-' + data['projection']
            header[f'CTYPE{idx2}'] = AXES_TO_CTYPE[data['system']][data['axes'][1]] + '-' + data['projection']
            header[f'CRPIX{idx1}'] = data['crpix'][0] + 1
            header[f'CRPIX{idx2}'] = data['crpix'][1] + 1
            header[f'CRVAL{idx1}'] = data['crval'][0]
            header[f'CRVAL{idx2}'] = data['crval'][1]
            header[f'CDELT{idx1}'] = data['cdelt'][0]
            header[f'CDELT{idx2}'] = data['cdelt'][1]
            header[f'CUNIT{idx1}'] = sanitize_unit(data['units'][0])
            header[f'CUNIT{idx2}'] = sanitize_unit(data['units'][1])
            header['LONPOLE'] = data['longpole']
            header['LATPOLE'] = data['latpole']
            if data.get('system') in EQUATORIAL_SYSTEMS:
                header['RADESYS'] = RADESYS[data['conversionSystem']]
                if data['conversionSystem'] in EQUINOX:
                    header['EQUINOX'] = EQUINOX[data['conversionSystem']]
            # NOTE: unclear if it is deliberate that the following is always
            # ?_2 and ?_1 or whether it should depend on the index of the
            # longitude.
            if data.get('system') in EQUATORIAL_SYSTEMS:
                header[f'PV{idx2}_1'] = 0.
                header[f'PV{idx2}_2'] = 0.
            header[f'PC{idx1}_{idx1}'] = data['pc'][0, 0]
            header[f'PC{idx1}_{idx2}'] = data['pc'][0, 1]
            header[f'PC{idx2}_{idx1}'] = data['pc'][1, 0]
            header[f'PC{idx2}_{idx2}'] = data['pc'][1, 1]
        elif coord_type == 'stokes':
            idx = worldmap[index][0] + 1
            header[f'CTYPE{idx}'] = 'STOKES'
            header[f'CRVAL{idx}'] = data['crval'][0]
            header[f'CRPIX{idx}'] = data['crpix'][0] + 1
            header[f'CDELT{idx}'] = data['cdelt'][0]
            header[f'CUNIT{idx}'] = ''
            header[f'PC{idx}_{idx}'] = data['pc'][0][0]
        elif coord_type == 'spectral':
            idx = worldmap[index][0] + 1
            if 'tabular' in data:
                header[f'CTYPE{idx}'] = AXES_TO_CTYPE[data['tabular']['axes'][0]]
                header[f'CRVAL{idx}'] = data['tabular']['crval'][0]
                header[f'CRPIX{idx}'] = data['tabular']['crpix'][0] + 1
                # It looks like we can't just use:
                # header[f'CDELT{idx}'] = data['tabular']['cdelt'][0]
                # because this doesn't match what appears in a FITS header
                # exported from CASA. Instead we use the interval between the
                # first two tabulated values. See
                # https://github.com/radio-astro-tools/spectral-cube/issues/614
                # for more information.
                header[f'CDELT{idx}'] = np.diff(data['tabular']['worldvalues'][:2])[0]
                header[f'CUNIT{idx}'] = data['tabular']['units'][0]
            else:
                header[f'CTYPE{idx}'] = data['wcs']['ctype']
                header[f'CRVAL{idx}'] = data['wcs']['crval']
                header[f'CRPIX{idx}'] = data['wcs']['crpix'] + 1
                header[f'CDELT{idx}'] = data['wcs']['cdelt']
                header[f'CUNIT{idx}'] = data['unit']
            header[f'PC{idx}_{idx}'] = 1.0
            header['RESTFRQ'] = data['restfreq']
            if data['system'] in SPECSYS:
                header['SPECSYS'] = SPECSYS[data['system']]
        elif coord_type == 'linear':
            indices = worldmap[index] + 1
            for idx1 in range(len(indices)):
                header[f'CTYPE{indices[idx1]}'] = data['axes'][idx1].upper()
                header[f'CRVAL{indices[idx1]}'] = data['crval'][idx1]
                header[f'CRPIX{indices[idx1]}'] = data['crpix'][idx1] + 1
                header[f'CDELT{indices[idx1]}'] = data['cdelt'][idx1]
                header[f'CUNIT{indices[idx1]}'] = data['units'][idx1]
                for idx2 in range(len(indices)):
                    header[f'PC{indices[idx1]}_{indices[idx1]}'] = data['pc'][idx1, idx1]
                    header[f'PC{indices[idx1]}_{indices[idx2]}'] = data['pc'][idx1, idx2]
                    header[f'PC{indices[idx2]}_{indices[idx1]}'] = data['pc'][idx2, idx1]
                    header[f'PC{indices[idx2]}_{indices[idx2]}'] = data['pc'][idx2, idx2]
        else:
            raise NotImplementedError(f'coord_type is {coord_type}')

    return WCS(header)



from __future__ import print_function, absolute_import, division

import os
import glob
import pytest
import numpy as np
from numpy.testing import assert_equal
from astropy.table import Table
from pprint import pformat

from ..data_managers import TiledCellStMan
from ..casa_functions import getdminfo, getdesc
from ..core import EndianAwareFileHandle
from ..table import CASATable

try:
    from casatools import table, image
    CASATOOLS_INSTALLED = True
except ImportError:
    CASATOOLS_INSTALLED = False

DATA = os.path.join(os.path.dirname(__file__), 'data')


SHAPES = [(3,), (5, 3), (8, 4, 2), (4, 8, 3, 1), (133, 400), (100, 211, 201),
          (50, 61, 72, 83), (4, 8, 10, 20, 40)]


@pytest.mark.skipif('not CASATOOLS_INSTALLED')
@pytest.mark.parametrize('shape', SHAPES)
def test_getdminfo(tmp_path, shape):

    filename = str(tmp_path / 'test.image')

    data = np.random.random(shape)

    ia = image()
    ia.fromarray(outfile=filename, pixels=data, log=False)
    ia.close()

    tb = table()
    tb.open(filename)
    reference = tb.getdminfo()
    tb.close()

    actual = getdminfo(filename)

    # The easiest way to compare the output is simply to compare the output
    # from pformat (checking for dictionary equality doesn't work because of
    # the Numpy arrays inside). Older versions of casatools represent some of
    # the vectors as int32 instead of native int but our implementation uses
    # native int so strip any mention of int32 from reference output
    assert pformat(actual) == pformat(reference).replace(', dtype=int32', '').replace('U3', 'U16')


def test_getdminfo_large():

    # Check that things continue to work fine when we cross the threshold from
    # a dataset with a size that can be represented by a 32-bit integer to one
    # where the size requires a 64-bit integer. We use pre-generated
    # table.f0 files here since generating these kinds of datasets is otherwise
    # slow and consumes a lot of memory.

    filename = os.path.join(DATA, 'lt32bit.image')
    with open(os.path.join(filename, 'table.f0'), 'rb') as f_orig:
        f = EndianAwareFileHandle(f_orig, '>', filename)
        magic = f.read(4)  # noqa
        lt32bit = TiledCellStMan()
        lt32bit.read_header(f)
    assert_equal(lt32bit.cube_shapes[0], (320, 320, 1, 1920))

    filename = os.path.join(DATA, 'gt32bit.image')
    with open(os.path.join(filename, 'table.f0'), 'rb') as f_orig:
        f = EndianAwareFileHandle(f_orig, '>', filename)
        magic = f.read(4)  # noqa
        gt32bit = TiledCellStMan()
        gt32bit.read_header(f)
    assert_equal(gt32bit.cube_shapes[0], (640, 640, 1, 1920))


@pytest.fixture
def filename(request):
    return request.getfixturevalue(request.param)


@pytest.mark.skipif('not CASATOOLS_INSTALLED')
def test_generic_table_read(tmp_path):

    # NOTE: for now, this doesn't check that we can read the data - just
    # the metadata about the table.

    filename_fits = str(tmp_path / 'generic.fits')
    filename_casa = str(tmp_path / 'generic.image')

    N = 120

    t = Table()
    t['short'] = np.arange(N, dtype=np.int16)
    t['ushort'] = np.arange(N, dtype=np.uint16)
    t['int'] = np.arange(N, dtype=np.int32)
    t['uint'] = np.arange(N, dtype=np.uint32)
    t['float'] = np.arange(N, dtype=np.float32)
    t['double'] = np.arange(N, dtype=np.float64)
    t['complex'] = np.array([1 + 2j, 3.3 + 8.2j, -1.2 - 4.2j] * (N // 3), dtype=np.complex64)
    t['dcomplex'] = np.array([3.33 + 4.22j, 3.3 + 8.2j, -1.2 - 4.2j] * (N // 3), dtype=np.complex128)
    t['str'] = np.array(['reading', 'casa', 'images'] * (N // 3))

    # Repeat this at the end to make sure we correctly finished reading
    # the complex column metadata
    t['int2'] = np.arange(N, dtype=np.int32)

    t.write(filename_fits)

    tb = table()
    tb.fromfits(filename_casa, filename_fits)
    tb.close()

    # Use the arrays in the table to also generate keywords of various types
    keywords = {'scalars': {}, 'arrays': {}}
    for name in t.colnames:
        keywords['scalars']['s_' + name] = t[name][0]
        keywords['arrays']['a_' + name] = t[name]

    tb.open(filename_casa, nomodify=False)
    tb.putkeywords(keywords)
    tb.flush()
    tb.close()

    desc_actual = getdesc(filename_casa, endian='<')

    tb.open(filename_casa)
    desc_reference = tb.getdesc()
    reference_getdminfo = tb.getdminfo()
    tb.close()

    assert pformat(desc_actual).replace(', dtype=int32', '') == pformat(desc_reference).replace(', dtype=int32', '')

    actual_getdminfo = getdminfo(filename_casa, endian='<')

    assert pformat(actual_getdminfo).replace('<U16', '<U8') == pformat(reference_getdminfo)

    tnew = Table.read(filename_casa)


def test_getdesc_floatarray():

    # There doesn't seem to be an easy way to create CASA images
    # with float (not double) arrays. In test_getdesc, all the floats
    # end up getting converted to double. So instead we use a table.dat
    # file that was found in the wild.

    desc = getdesc(os.path.join(DATA, 'floatarray.image'))
    trc = desc['_keywords_']['masks']['mask0']['box']['trc']
    assert trc.dtype == np.dtype('<f4')
    assert_equal(trc, [512, 512, 1, 100])


@pytest.mark.skipif('not CASATOOLS_INSTALLED')
def test_logtable(tmp_path):

    filename = str(tmp_path / 'test.image')
    logtable = str(tmp_path / 'test.image' / 'logtable')

    data = np.random.random((2, 3, 4))

    ia = image()
    ia.fromarray(outfile=filename, pixels=data, log=False)
    ia.sethistory(origin='test', history=['a', 'bb', 'ccccccccccc'] * 3247)
    ia.close()

    tb = table()
    tb.open(logtable)
    reference_getdesc = tb.getdesc()
    reference_getdminfo = tb.getdminfo()
    tb.close()

    actual_getdesc = getdesc(logtable, endian='<')
    actual_getdminfo = getdminfo(logtable, endian='<')

    assert pformat(actual_getdesc) == pformat(reference_getdesc)
    assert pformat(actual_getdminfo) == pformat(reference_getdminfo)

    Table.read(logtable)


@pytest.mark.parametrize('tablename', ('.',
                                       'ANTENNA',
                                       'CALDEVICE',
                                       'DATA_DESCRIPTION',
                                       'FEED',
                                       'FIELD',
                                       'FLAG_CMD',
                                       'HISTORY',
                                       'OBSERVATION',
                                       'POINTING',
                                       'POLARIZATION',
                                       'PROCESSOR',
                                       'SOURCE',
                                       'SPECTRAL_WINDOW',
                                       'STATE',
                                       'SYSCAL',
                                       'SYSPOWER'))
def test_ms_tables(tablename):

    table_filename = os.path.join(DATA, 'simple.ms', tablename)

    # Concatenation issue as arrays change shape half way through
    if tablename == 'SYSPOWER':
        pytest.xfail()

    tt = [Table.read(table_filename, data_desc_id=0),
          Table.read(table_filename, data_desc_id=1)]

    if CASATOOLS_INSTALLED:

        tb = table()
        tb.open(table_filename)

        for colname in tb.colnames():

            # CASA has issues reading this in
            if tablename == 'SOURCE' and colname in ['POSITION', 'TRANSITION']:
                continue
            if tablename == 'CALDEVICE' and colname in ['CAL_EFF', 'TEMPERATURE_LOAD']:
                continue
            if tablename == 'SPECTRAL_WINDOW' and colname in ['CHAN_FREQ', 'CHAN_WIDTH', 'EFFECTIVE_BW',
                                                              'RESOLUTION', 'ASSOC_SPW_ID', 'ASSOC_NATURE']:
                continue

            # Column is empty? browsetable also has issues reading this properly
            if tablename == '.' and colname in ['FLAG_CATEGORY']:
                continue

            if tablename == '.':
                # Check two main sections of table, which have different shapes for
                # some of the columns
                assert_equal(tt[0][colname], tb.getcol(colname, startrow=0, nrow=10).T)
                assert_equal(tt[1][colname], tb.getcol(colname, startrow=10, nrow=10).T)
            else:
                assert_equal(tt[0][colname], tb.getcol(colname).T)

        tb.close()


@pytest.mark.skipif('not CASATOOLS_INSTALLED')
def test_vector_columns(tmp_path):

    filename_fits = str(tmp_path / 'vector.fits')
    filename_casa = str(tmp_path / 'vector.image')

    N = 120

    t = Table()
    t['short'] = np.arange(N, dtype=np.int16).reshape((5, 4, 3, 2))
    t['int'] = np.arange(N, dtype=np.int32).reshape((5, 4, 3, 2))
    t['double'] = np.arange(N, dtype=np.float64).reshape((5, 4, 3, 2))
    t.write(filename_fits)

    tb = table()
    tb.fromfits(filename_casa, filename_fits)
    tb.close()

    t2 = Table.read(filename_casa)

    assert_equal(t['short'], t2['short'])
    assert_equal(t['int'], t2['int'])
    assert_equal(t['double'], t2['double'])


# Test reading tables in the casadata package

try:
    from casadata import datapath
except ImportError:
    CASADATA_TABLES = []
    CASADATA_INSTALLED = True
else:
    CASADATA_TABLES = [os.path.dirname(x) for x in glob.glob(os.path.join(datapath, '**', 'table.dat'), recursive=True)]
    CASADATA_INSTALLED = False


EXPECTED_FAILURES = [
    'ephemerides/Sources',  # our table is too long
]


@pytest.mark.parametrize('table_filename', CASADATA_TABLES)
def test_casadata(table_filename):

    if os.path.relpath(table_filename, datapath) in EXPECTED_FAILURES:
        pytest.xfail()

    tt = Table.read(table_filename)

    if CASATOOLS_INSTALLED:

        tb = table()
        tb.open(table_filename)

        expected_colnames = tb.colnames()

        # Some tables contain a Spectral_Record column that we can't read yet
        # but that browsetable also can't read/display so we don't support this.
        if 'Spectral_Record' in expected_colnames:
            expected_colnames.remove('Spectral_Record')

        assert tt.colnames == expected_colnames

        for colname in expected_colnames:
            try:
                expected_data = tb.getcol(colname).T
            except Exception:
                # If casa can't read it, we can assume we don't need to support it
                continue
            assert_equal(tt[colname], expected_data)

        tb.close()



import os
from astropy.table import Table
from astropy.io import registry

from casa_formats_io.casa_low_level_io.table import CASATable


def identify_casa_table(origin, *args, **kwargs):
    if (isinstance(args[2], str) and
            os.path.isdir(args[2]) and
            os.path.exists(os.path.join(args[2], 'table.dat'))):
        with open(os.path.join(args[2], 'table.dat'), 'rb') as f:
            return f.read(4) == b'\xbe\xbe\xbe\xbe'


def read_casa_table(filename, data_desc_id=None):
    table = CASATable.read(filename)
    return table.as_astropy_table(data_desc_id=data_desc_id)


registry.register_identifier('casa-table', Table, identify_casa_table)
registry.register_reader('casa-table', Table, read_casa_table)



import os
from math import prod
from collections import defaultdict

import numpy as np

import dask.array as da

from ..core import (check_type_and_version, BaseCasaObject, with_nbytes_prefix,
                    read_string, read_int32, read_int64, read_iposition,
                    TO_DTYPE, Block)

from ..record import Record

__all__ = ['VariableShapeArrayList', 'TiledStMan', 'TiledCellStMan',
           'TiledShapeStMan', 'TiledColumnStMan']


class VariableShapeArrayList(list):
    pass


class TiledStMan(BaseCasaObject):

    @with_nbytes_prefix
    def read_header(self, f):

        version = check_type_and_version(f, 'TiledStMan', (1, 2))

        if version >= 2:
            self.big_endian = f.read(1) == b'\x01'  # noqa
        else:
            self.big_endian = True

        # TODO: Set endian flag on f here

        self.seqnr = read_int32(f)
        # if self.seqnr != 0:
        #     raise ValueError("Expected seqnr to be 0, got {0}".format(self.seqnr))

        self.nrows = read_int32(f)
        # if self.nrows != 1:
        #     raise ValueError("Expected nrows to be 1, got {0}".format(self.nrows))

        self.ncols = read_int32(f)
        if self.ncols != 1:
            raise ValueError("Expected ncols to be 1, got {0}".format(self.ncols))

        self.dtype = read_int32(f)
        self.column_name = read_string(f)
        self.max_cache_size = read_int32(f)
        self.ndim = read_int32(f)

        self.nrfile = read_int32(f)  # 1
        # if self.nrfile != 1:
        #     raise ValueError("Expected nrfile to be 1, got {0}".format(self.nrfile))

        self.max_tsm_index = 0

        for itsm in range(self.nrfile):

            # The following flag seems to control whether or not the TSM file is
            # opened by CASA, and is probably safe to ignore here.
            flag = bool(f.read(1) == b'\x01')

            if not flag:
                continue

            self.max_tsm_index = itsm

            # The following two values are unknown, but are likely relevant when there
            # are more that one field in the image.

            mode = read_int32(f)
            unknown = read_int32(f)  # 0

            if mode == 1:
                self.total_cube_size = read_int32(f)
            elif mode == 2:
                self.total_cube_size = read_int64(f)
            else:
                raise ValueError('Unexpected value {0} at position {1}'.format(mode, f.tell() - 8))

        unknown = read_int32(f)  # 1

        self.cube_shapes = []
        self.tile_shapes = []

        for itsm in range(self.nrfile):

            unknown = read_int32(f)  # 1

            Record.read(f)

            flag = f.read(1)  # noqa

            ndim2 = read_int32(f)  # noqa

            self.cube_shapes.append(read_iposition(f))
            self.tile_shapes.append(read_iposition(f))

            unknown = read_int32(f)  # noqa
            unknown = read_int32(f)  # noqa

    def read_column(self, filename, seqnr, column, coldesc, colindex_in_dm):

        # chunkshape defines how the chunks (array subsets) are written to disk
        chunkshape = tuple(self.tile_shapes[0])

        # the total shape defines the final output array shape
        if len(self.cube_shapes[0]) > 0:
            totalshape = self.cube_shapes[0]
        else:
            # FIXME: below is not the right default!
            totalshape = np.array(chunkshape)

        return self._read_tsm_file(filename, seqnr, coldesc, totalshape, chunkshape)

    def _read_tsm_file(self, filename, seqnr, coldesc, totalshape, chunkshape, tsm_index=0, offset=0):

        totalshape = np.asarray(totalshape)
        chunkshape = np.asarray(chunkshape)
        chunksize = prod(chunkshape)

        # Need to expose the following somehow
        target_chunksize = None
        memmap = True

        # the ratio between these tells you how many chunks must be combined
        # to create a final stack along each dimension
        stacks = np.ceil(totalshape / chunkshape).astype(int)

        dtype = TO_DTYPE[coldesc.value_type]

        if coldesc.value_type != 'bool':
            if self.big_endian:
                dtype = '>' + dtype
            else:
                dtype = '<' + dtype

        itemsize = np.dtype(dtype).itemsize

        chunkshape = tuple(int(x) for x in chunkshape)
        totalshape = tuple(int(x) for x in totalshape)

        # CASA chunks are typically too small to be efficient, so we use a larger
        # chunk size for dask and then tell CASAArrayWrapper about both the native
        # and target chunk size.
        # chunkshape = determine_optimal_chunkshape(totalshape, chunkshape)

        if target_chunksize is None:
            target_chunksize = 10000000

        if chunksize < target_chunksize:

            # Find optimal chunk - since we want to be efficient we want the new
            # chunks to be contiguous on disk so we first try and increase the
            # chunk size in x, then y, etc.

            chunkoversample = previous_chunkoversample = [1 for i in range(len(chunkshape))]

            finished = False
            for dim in range(len(chunkshape)):
                factors = [f for f in range(stacks[dim] + 1) if stacks[dim] % f == 0]
                for factor in factors:
                    chunkoversample[dim] = factor
                    if prod(chunkoversample) * chunksize > target_chunksize:
                        chunkoversample = previous_chunkoversample
                        finished = True
                        break
                    previous_chunkoversample = chunkoversample.copy()
                if finished:
                    break

        else:

            chunkoversample = (1,) * len(chunkshape)

        chunkshape = [c * o for (c, o) in zip(chunkshape, chunkoversample)]

        # Create a wrapper that takes slices and returns the appropriate CASA data
        from casa_formats_io.casa_dask import CASAArrayWrapper

        img_fn = os.path.join(filename, f'table.f{seqnr}_TSM{tsm_index}')

        wrapper = CASAArrayWrapper(img_fn, totalshape, chunkshape,
                                   chunkoversample=chunkoversample, dtype=dtype,
                                   itemsize=itemsize, memmap=memmap, offset=offset)

        # Convert to a dask array
        import uuid
        from dask.array import from_array
        dask_array = from_array(wrapper, name='CASA Data ' + str(uuid.uuid4()),
                                chunks=chunkshape[::-1], meta=np.array([[[]]], dtype=dtype))

        # Since the chunks may not divide the array exactly, all the chunks put
        # together may be larger than the array, so we need to get rid of any
        # extraneous padding.
        final_slice = tuple([slice(dim) for dim in totalshape[::-1]])

        return dask_array[final_slice]


class TiledCellStMan(TiledStMan):

    @classmethod
    def read(cls, f):
        self = cls()
        self.name = 'TiledCellStMan'
        return self

    @with_nbytes_prefix
    def read_header(self, f):

        # The code in this function corresponds to TiledStMan::headerFileGet
        # https://github.com/casacore/casacore/blob/75b358be47039250e03e5042210cbc60beaaf6e4/tables/DataMan/TiledStMan.cc#L1086

        check_type_and_version(f, 'TiledCellStMan', 1)

        self.default_tile_shape = read_iposition(f)

        super(TiledCellStMan, self).read_header(f)

    def read_column(self, *args, **kwargs):
        array = super(TiledCellStMan, self).read_column(*args, **kwargs)
        array = array.reshape((1,) + array.shape)
        return array


class TiledShapeStMan(TiledStMan):

    @classmethod
    def read(cls, f):
        self = cls()
        self.name = 'TiledShapeStMan'
        return self

    @with_nbytes_prefix
    def read_header(self, f):
        check_type_and_version(f, 'TiledShapeStMan', 1)
        super(TiledShapeStMan, self).read_header(f)
        self.default_tile_shape = read_iposition(f)
        self.number_used_row_map = read_int32(f)

        # The data might be split into multiple cubes (TSM<index> files). The
        # following three values help us piece these together into the main
        # hypercube. Each of the lists has length n_cubes where n_cubes is the
        # number of individual cubes.

        # The index of the last row in the final hypercube in each section. For
        # instance, [9, 19] means that the ten first rows (0-9) are in the first
        # subcube and the second set of ten rows (10-19) are in the second cube.
        self.last_row_abs = Block.read(f, read_int32)

        # The index of the cube in which the rows are stored - this is the value
        # used as a suffix in the TSM filename, e.g. TSM2
        self.cube_index = Block.read(f, read_int32)

        # The index of the last row of the subcube.
        self.last_row_sub = Block.read(f, read_int32)

    def read_column(self, filename, seqnr, column, coldesc, colindex_in_dm):

        # chunkshape defines how the chunks (array subsets) are written to disk
        chunkshape = list(self.default_tile_shape)

        if len(self.last_row_abs.elements) == 0:
            return VariableShapeArrayList([])

        tsm_indices = np.unique(self.cube_index.elements)

        # Start off by reading each TSM file into a dask array
        dask_arrays = {}
        for itsm in tsm_indices:
            dask_arrays[itsm] = self._read_tsm_file(filename, seqnr, coldesc,
                                                    self.cube_shapes[itsm],
                                                    self.tile_shapes[itsm],
                                                    tsm_index=itsm)

        # Next up, we construct for each of the hypercubes an array of row indices
        # in the final table that each hypercube row corresponds to.

        # Start off by making a master index of all rows
        index = da.arange(self.last_row_abs.elements[-1] + 1)

        # Construct the index for each TSM cube
        tsm_row_index = defaultdict(list)
        for idx in range(len(self.cube_index.elements)):
            start = 0 if idx == 0 else self.last_row_abs.elements[idx - 1] + 1
            end = self.last_row_abs.elements[idx] + 1
            tsm_row_index[self.cube_index.elements[idx]].append(index[start:end])

        # Make each row index into a single dask array
        for itsm in tsm_row_index:
            tsm_row_index[itsm] = da.hstack(tsm_row_index[itsm])

        # Return a list of (row_index, hypercube) tuples
        return VariableShapeArrayList([(tsm_row_index[itsm], dask_arrays[itsm]) for itsm in tsm_indices])


class TiledColumnStMan(TiledStMan):

    @classmethod
    def read(cls, f):
        self = cls()
        self.name = 'TiledColumnStMan'
        return self

    @with_nbytes_prefix
    def read_header(self, f):
        check_type_and_version(f, 'TiledColumnStMan', 1)
        self.default_tile_shape = read_iposition(f)
        super(TiledColumnStMan, self).read_header(f)



from __future__ import print_function, absolute_import, division

import os
from itertools import product

import pytest
import numpy as np
from numpy.testing import assert_allclose

from ..casa_dask import image_to_dask

try:
    from casatools import image
    CASA_INSTALLED = True
except ImportError:
    try:
        from taskinit import ia as image
        CASA_INSTALLED = True
    except ImportError:
        CASA_INSTALLED = False

# NOTE: the (127, 337, 109) example is to make sure that things work correctly
# when the shape isn't a multiple of the chunk size along any
# dimension.
SHAPES = [(3, 4, 5), (129, 128, 130), (513, 128, 128), (128, 513, 128),
          (128, 128, 513), (512, 64, 64), (127, 337, 109)]


@pytest.mark.skipif(not CASA_INSTALLED, reason='CASA tests must be run in a CASA environment.')
@pytest.mark.parametrize(('memmap', 'shape'), product([False, True], SHAPES))
def test_image_to_dask(tmpdir, memmap, shape):

    # Unit tests for the low-level image_to_dask function which can
    # read a CASA image or mask to a Dask array.

    reference = np.random.random(shape).astype(np.float32)

    # CASA seems to have precision issues when computing masks with values
    # very close to e.g. 0.5 in >0.5. To avoid this, we filter out random
    # values close to the boundaries that we use below.
    reference[np.isclose(reference, 0.2)] += 0.05
    reference[np.isclose(reference, 0.5)] += 0.05
    reference[np.isclose(reference, 0.8)] += 0.05

    os.chdir(tmpdir.strpath)

    # Start off with a simple example with no mask. Note that CASA requires
    # the array to be transposed in order to match what we would expect.

    ia = image()
    ia.fromarray('basic.image', pixels=reference.T, log=False)
    ia.close()

    array1 = image_to_dask('basic.image', memmap=memmap)
    assert array1.dtype == np.float32
    assert_allclose(array1, reference)

    # Check slicing
    assert_allclose(array1[:2, :1, :3], reference[:2, :1, :3])

    # Try and get a mask - this should fail since there isn't one.

    with pytest.raises(FileNotFoundError):
        image_to_dask('basic.image', mask=True, memmap=memmap)

    # Now create an array with a simple uniform mask.

    ia = image()
    ia.fromarray('scalar_mask.image', pixels=reference.T, log=False)
    ia.calcmask(mask='T')
    ia.close()

    array2 = image_to_dask('scalar_mask.image', memmap=memmap)
    assert_allclose(array2, reference)

    mask2 = image_to_dask('scalar_mask.image', mask=True, memmap=memmap)
    assert mask2.dtype is np.dtype('bool')
    assert mask2.shape == array2.shape
    assert np.all(mask2)

    # Check with a full 3-d mask

    ia = image()
    ia.fromarray('array_mask.image', pixels=reference.T, log=False)
    ia.calcmask(mask='array_mask.image>0.5')
    ia.close()

    array3 = image_to_dask('array_mask.image', memmap=memmap)
    assert_allclose(array3, reference)

    mask3 = image_to_dask('array_mask.image', mask=True, memmap=memmap)
    assert_allclose(mask3, reference > 0.5)

    # Check slicing
    assert_allclose(mask3[:2, :1, :3], (reference > 0.5)[:2, :1, :3])

    # Test specifying the mask name

    ia = image()
    ia.fromarray('array_masks.image', pixels=reference.T, log=False)
    ia.calcmask(mask='array_masks.image>0.5')
    ia.calcmask(mask='array_masks.image>0.2')
    ia.calcmask(mask='array_masks.image>0.8', name='gt08')
    ia.close()

    array4 = image_to_dask('array_masks.image', memmap=memmap)
    assert_allclose(array4, reference)

    mask4 = image_to_dask('array_masks.image', mask=True, memmap=memmap)
    assert_allclose(mask4, reference > 0.5)

    mask5 = image_to_dask('array_masks.image', mask='mask0', memmap=memmap)
    assert_allclose(mask5, reference > 0.5)

    mask6 = image_to_dask('array_masks.image', mask='mask1', memmap=memmap)
    assert_allclose(mask6, reference > 0.2)

    mask7 = image_to_dask('array_masks.image', mask='gt08', memmap=memmap)
    assert_allclose(mask7, reference > 0.8)

    # Check that things still work if we write the array out with doubles

    reference = np.random.random(shape).astype(np.float64)

    ia = image()
    ia.fromarray('double.image', pixels=reference.T, type='d', log=False)
    ia.close()

    array8 = image_to_dask('double.image', memmap=memmap)
    assert array8.dtype == np.float64
    assert_allclose(array8, reference)


@pytest.mark.skipif(not CASA_INSTALLED, reason='CASA tests must be run in a CASA environment.')
def test_target_chunksize():

    reference = np.random.random((256, 256, 256))

    ia = image()
    ia.fromarray('large.image', pixels=reference, log=False)
    ia.close()

    array1 = image_to_dask('large.image')
    assert array1.chunksize == (128, 256, 256)

    array2 = image_to_dask('large.image', target_chunksize=100000)
    assert array2.chunksize == (32, 32, 64)

    array3 = image_to_dask('large.image', target_chunksize=1000)
    assert array3.chunksize == (32, 32, 32)



from __future__ import print_function, absolute_import, division

import os
import tempfile

import pytest
import numpy as np
from astropy.wcs import WCS
from astropy.io import fits

from numpy.testing import assert_allclose

from ..casa_low_level_io.table import CASATable
from ..casa_wcs import coordsys_to_astropy_wcs

HEADER_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'header_jybeam.hdr')

try:
    from casatools import image
    CASATOOLS_INSTALLED = True
except ImportError:
    CASATOOLS_INSTALLED = False


@pytest.fixture
def filename(request):
    return request.getfixturevalue(request.param)


def assert_header_correct(casa_filename):

    fits_filename = tempfile.mktemp()

    # Use CASA to convert back to FITS and use that header as the reference

    ia = image()
    ia.open(casa_filename)
    ia.tofits(fits_filename, stokeslast=False)
    ia.done()
    ia.close()

    # Parse header with WCS - for the purposes of this function
    # we are not interested in keywords/values not in WCS
    reference_wcs = WCS(fits_filename)
    reference_header = reference_wcs.to_header()

    # Now use our coordsys_to_astropy_wcs function to create the header and compare
    # the results.
    table = CASATable.read(casa_filename, endian='>')
    coords = table.desc.keywords.as_dict()['coords']
    actual_wcs = coordsys_to_astropy_wcs(coords)
    actual_header = actual_wcs.to_header()

    assert sorted(actual_header) == sorted(reference_header)

    for key in reference_header:
        if isinstance(actual_header[key], str):
            assert actual_header[key] == reference_header[key]
        else:
            assert_allclose(actual_header[key], reference_header[key])


@pytest.mark.skipif('not CASATOOLS_INSTALLED')
def test_coordsys_to_astropy_wcs_linear(tmp_path):

    # Test that things work properly when the WCS coordinates aren't set

    casa_filename = str(tmp_path / 'test.image')

    data = np.random.random((3, 4, 5, 6, 7))

    ia = image()
    ia.fromarray(outfile=casa_filename, pixels=data, log=False)
    ia.close()

    assert_header_correct(casa_filename)


def header_copy_with(**kwargs):
    header = fits.Header.fromtextfile(HEADER_FILENAME).copy()
    header.update(kwargs)
    return header


ALL_HEADERS = [
    header_copy_with(),
    header_copy_with(CTYPE1='GLON-TAN', CTYPE2='GLAT-TAN'),
    header_copy_with(CTYPE1='SLON-TAN', CTYPE2='SLAT-TAN'),
    header_copy_with(CTYPE1='ELON-TAN', CTYPE2='ELAT-TAN'),
    header_copy_with(CTYPE1='HLON-TAN', CTYPE2='HLAT-TAN'),
    header_copy_with(SPECSYS=''),
    header_copy_with(SPECSYS='TOPOCENT'),
    header_copy_with(SPECSYS='GEOCENTR'),
    header_copy_with(SPECSYS='BARYCENT'),
    header_copy_with(SPECSYS='HELIOCEN'),
    header_copy_with(SPECSYS='LSRK'),
    header_copy_with(SPECSYS='LSRD'),
    header_copy_with(SPECSYS='GALACTOC'),
    header_copy_with(SPECSYS='LOCALGRP'),
    header_copy_with(SPECSYS='CMBDIPOL'),
    header_copy_with(SPECSYS='SOURCE'),
    header_copy_with(RADESYS='FK4'),
    header_copy_with(RADESYS='FK4-NO-E'),
    header_copy_with(RADESYS='FK5'),
    header_copy_with(RADESYS='ICRS'),
    header_copy_with(EQUINOX=1950.),
    header_copy_with(EQUINOX=1979.9),
    header_copy_with(EQUINOX=2000),
    header_copy_with(EQUINOX=2010),
    header_copy_with(CTYPE3='FREQ', CUNIT3='GHz', CRVAL3=100., CDELT3=1.),
    header_copy_with(CTYPE3='WAVE', CUNIT3='m', CRVAL3=1e-6, CDELT3=1e-8),
    header_copy_with(CTYPE3='VOPT'),
    header_copy_with(CTYPE3='VRAD')
]


@pytest.mark.skipif('not CASATOOLS_INSTALLED')
@pytest.mark.parametrize('header', ALL_HEADERS)
def test_coordsys_to_astropy_wcs_additional(tmp_path, header):

    # More cases to improve coverage

    casa_filename = str(tmp_path / 'casa.image')
    fits_filename = str(tmp_path / 'casa.fits')

    fits.writeto(fits_filename, np.ones((2, 3, 4, 5)), header)

    ia = image()
    ia.fromfits(infile=fits_filename, outfile=casa_filename)
    ia.unlock()
    ia.close()
    ia.done()

    assert_header_correct(casa_filename)



# Numpy + Dask implementation of CASA image data access

from __future__ import print_function, absolute_import, division

from math import prod
import os
from math import ceil
import uuid
import numpy as np

import dask.array

from .casa_low_level_io.table import CASATable
from ._casa_chunking import _combine_chunks

__all__ = ['image_to_dask']


def combine_chunks(array_1d, itemsize, shape, oversample):
    if len(shape) < 4:
        shape = tuple(shape) + (1,) * (4 - len(shape))
    if len(oversample) < 4:
        oversample = tuple(oversample) + (1,) * (4 - len(oversample))
    native_shape = [s // o for (s, o) in zip(shape, oversample)]
    return _combine_chunks(np.ascontiguousarray(array_1d), itemsize, *native_shape[::-1], *oversample[::-1])


class CASAArrayWrapper:
    """
    A wrapper class for dask that accesses chunks from a CASA file on request.
    It is assumed that this wrapper will be used to construct a dask array that
    has chunks aligned with the CASA file chunks.

    Having a single wrapper object such as this is far more efficient than
    having one array wrapper per chunk. This is because the dask graph gets
    very large if we end up with one dask array per chunk and slows everything
    down.
    """

    def __init__(self, filename, totalshape, chunkshape, chunkoversample=None,
                 dtype=None, itemsize=None, memmap=False, offset=0):
        self._filename = filename
        self._totalshape = totalshape[::-1]
        self._chunkshape = chunkshape[::-1]
        self._chunkoversample = chunkoversample[::-1]
        self.shape = totalshape[::-1]
        self.dtype = dtype
        self.ndim = len(self.shape)
        self._stacks = np.ceil(np.array(totalshape) / np.array(chunkshape)).astype(int)
        self._chunksize = prod(chunkshape)
        self._itemsize = itemsize
        self._memmap = memmap
        self._offset = offset
        if not memmap:
            if self._itemsize == 1:
                self._array = np.unpackbits(np.fromfile(filename, dtype='uint8'), bitorder='little')
            else:
                self._array = np.fromfile(filename, dtype=np.uint8)
        self._last_item = None
        self._last_result = None

    def __getitem__(self, item):

        # dask does not cache calls to __getitem__ so in some cases might call it twice with
        # the same item - rather than try and cache multiple possible items we implement a
        # simple caching strategy of at least ensuring that successive calls with the same
        # input are cached (https://github.com/dask/dask/issues/8420).
        # This is important for example for Table.__repr__ which accesses the cells
        # one by one.
        if item == self._last_item:
            return self._last_result

        # TODO: potentially normalize item, for now assume it is a list of slice objects

        indices = []
        for dim in range(self.ndim):
            if isinstance(item[dim], slice):
                indices.append(item[dim].start // self._chunkshape[dim])
            else:
                indices.append(item[dim] // self._chunkshape[dim])

        chunk_number = indices[0]
        for dim in range(1, self.ndim):
            chunk_number = chunk_number * self._stacks[::-1][dim] + indices[dim]

        offset = chunk_number * self._chunksize * self._itemsize + self._offset * self._itemsize

        item_in_chunk = []
        for dim in range(self.ndim):
            if isinstance(item[dim], slice):
                item_in_chunk.append(slice(item[dim].start - indices[dim] * self._chunkshape[dim],
                                           item[dim].stop - indices[dim] * self._chunkshape[dim],
                                           item[dim].step))
            else:
                item_in_chunk.append(item[dim] - indices[dim] * self._chunkshape[dim])
        item_in_chunk = tuple(item_in_chunk)

        if self._itemsize == 1:

            if self._memmap:
                n_native = prod(self._chunkoversample)
                rounded_up_chunksize = ceil(self._chunksize / n_native / 8) * n_native
                offset = offset // self._chunksize * rounded_up_chunksize
                array_uint8 = np.fromfile(self._filename, dtype=np.uint8,
                                          offset=offset, count=rounded_up_chunksize)
                array_bits = np.unpackbits(array_uint8, bitorder='little')
            else:
                array_bits = self._array[offset * 8: (offset + self._chunksize) * 8]

            chunk = combine_chunks(array_bits, 1,
                                   shape=self._chunkshape,
                                   oversample=self._chunkoversample)[:self._chunksize]

            result = chunk.reshape(self._chunkshape[::-1], order='F').T[item_in_chunk].astype(np.bool_)

        else:

            if self._memmap:
                data_bytes = np.fromfile(self._filename, dtype=np.uint8,
                                         offset=offset,
                                         count=self._chunksize * self._itemsize)
            else:
                data_bytes = self._array[chunk_number*self._chunksize * self._itemsize:
                                         (chunk_number + 1)*self._chunksize * self._itemsize]

            result =  (combine_chunks(data_bytes,
                                      self._itemsize,
                                      shape=self._chunkshape,
                                      oversample=self._chunkoversample)
                       .view(self.dtype)
                       .reshape(self._chunkshape[::-1], order='F').T[item_in_chunk])

        self._last_item = item
        self._last_result = result

        return result



def image_to_dask(imagename, memmap=True, mask=False, target_chunksize=None):
    """
    Read a CASA image (a folder containing a ``table.f0_TSM0`` file) into a
    dask array.

    Parameters
    ----------
    imagename : str
        The filename of the CASA image directory
    memmap : bool, optional
        Whether to use memory mapping or load the full cube into memory
    mask : str or bool, optional
        If set to a string, should be the name of the mask (e.g. ``mask1``) and
        the mask is returned instead of the data. If `True`, ``mask0`` will be used.
    target_chunksize : int
        The desired dask chunk size - CASA chunks are aggregated into blocks of
        approximately this number of elements.
    """

    # the data is stored in the following binary file
    # each of the chunks is stored on disk in fortran-order
    if mask:
        if mask is True:
            mask = 'mask0'
        imagename = os.path.join(str(imagename), mask)

    if not os.path.exists(imagename):
        raise FileNotFoundError(imagename)

    # the data is stored in the following binary file
    # each of the chunks is stored on disk in fortran-order
    img_fn = os.path.join(str(imagename), 'table.f0_TSM0')

    # load the metadata from the image table.
    table = CASATable.read(str(imagename), endian='>')

    # extract data manager
    dm = table.column_set.data_managers[0]

    # Determine whether file is big endian
    big_endian = dm.big_endian

    # chunkshape defines how the chunks (array subsets) are written to disk
    chunkshape = tuple(dm.default_tile_shape)
    chunksize = prod(chunkshape)

    # the total shape defines the final output array shape
    totalshape = dm.cube_shapes[0]

    # we expect that the total size of the array will be determined by finding
    # the number of chunks along each dimension rounded up
    totalsize = prod(np.ceil(totalshape / chunkshape)) * chunksize

    # the file size helps us figure out what the dtype of the array is
    filesize = os.stat(img_fn).st_size

    # the ratio between these tells you how many chunks must be combined
    # to create a final stack
    stacks = np.ceil(totalshape / chunkshape).astype(int)
    nchunks = int(prod(stacks))

    # check that the file size is as expected and determine the data dtype
    if mask:
        expected = nchunks * ceil(chunksize / 8)
        if filesize != expected:
            raise ValueError("Unexpected file size for mask, found {0} but "
                             "expected {1}".format(filesize, expected))
        dtype = bool
        itemsize = 1
    else:
        if filesize == totalsize * 4:
            if big_endian:
                dtype = '>f4'
            else:
                dtype = '<f4'
            itemsize = 4
        elif filesize == totalsize * 8:
            if big_endian:
                dtype = '>f8'
            else:
                dtype = '<f8'
            itemsize = 8
        else:
            raise ValueError("Unexpected file size for data, found {0} but "
                             "expected {1} or {2}".format(filesize, totalsize * 4, totalsize * 8))

    # CASA does not like numpy ints!
    chunkshape = tuple(int(x) for x in chunkshape)
    totalshape = tuple(int(x) for x in totalshape)

    # CASA chunks are typically too small to be efficient, so we use a larger
    # chunk size for dask and then tell CASAArrayWrapper about both the native
    # and target chunk size.
    # chunkshape = determine_optimal_chunkshape(totalshape, chunkshape)

    if target_chunksize is None:
        target_chunksize = 10000000

    if chunksize < target_chunksize:

        # Find optimal chunk - since we want to be efficient we want the new
        # chunks to be contiguous on disk so we first try and increase the
        # chunk size in x, then y, etc.

        chunkoversample = previous_chunkoversample = [1 for i in range(len(chunkshape))]

        finished = False
        for dim in range(len(chunkshape)):
            factors = [f for f in range(stacks[dim] + 1) if stacks[dim] % f == 0]
            for factor in factors:
                chunkoversample[dim] = factor
                if prod(chunkoversample) * chunksize > target_chunksize:
                    chunkoversample = previous_chunkoversample
                    finished = True
                    break
                previous_chunkoversample = chunkoversample.copy()
            if finished:
                break

    else:

        chunkoversample = (1,) * len(chunkshape)

    chunkshape = [c * o for (c, o) in zip(chunkshape, chunkoversample)]

    # Create a wrapper that takes slices and returns the appropriate CASA data
    wrapper = CASAArrayWrapper(img_fn, totalshape, chunkshape,
                               chunkoversample=chunkoversample, dtype=dtype,
                               itemsize=itemsize, memmap=memmap)

    # Convert to a dask array
    dask_array = dask.array.from_array(wrapper, name='CASA Data ' + str(uuid.uuid4()),
                                       chunks=chunkshape[::-1])

    # Since the chunks may not divide the array exactly, all the chunks put
    # together may be larger than the array, so we need to get rid of any
    # extraneous padding.
    final_slice = tuple([slice(dim) for dim in totalshape[::-1]])

    return dask_array[final_slice]



import os
from math import prod

import numpy as np

from ..core import (check_type_and_version, BaseCasaObject, with_nbytes_prefix,
                    read_string, read_int32, read_int64, read_as_numpy_array,
                    Block, EndianAwareFileHandle, TO_DTYPE)

__all__ = ['IncrementalStMan']


class ISMIndex(BaseCasaObject):

    # https://github.com/casacore/casacore/blob/dbf28794ef446bbf4e6150653dbe404379a3c429/tables/DataMan/ISMIndex.cc#L51s

    @classmethod
    @with_nbytes_prefix
    def read(cls, f):
        self = cls()
        version = check_type_and_version(f, 'ISMIndex', 1)
        self.n_used = read_int32(f)
        if version > 1:
            self.last_row = Block.read(f, read_int64)
        else:
            self.last_row = Block.read(f, read_int32)
        self.bucket_number = Block.read(f, read_int32)
        return self


class IncrementalStMan(BaseCasaObject):

    @classmethod
    @with_nbytes_prefix
    def read(cls, f):
        self = cls()
        check_type_and_version(f, 'ISM', 3)
        self.name = read_string(f)
        return self

    @with_nbytes_prefix
    def read_header(self, f):

        # SSMBase::readHeader()
        # https://github.com/casacore/casacore/blob/d6da19830fa470bdd8434fd855abe79037fda78c/tables/DataMan/SSMBase.cc#L415

        version = check_type_and_version(f, 'IncrementalStMan', (4, 5))

        if version >= 5:
            self.big_endian = f.read(1) == b'\x01'  # noqa
        else:
            self.big_endian = True

        self.bucket_size = read_int32(f)
        self.number_of_buckets = read_int32(f)
        self.persistent_cache = read_int32(f)

        # Unique nr for column in this storage manager
        self.unique_number_column = read_int32(f)

        if version > 1:
            self.number_of_free_buckets = read_int32(f)
            self.first_free_bucket = read_int32(f)

    def read_column(self, filename, seqnr, column, coldesc, colindex_in_dm):

        # Open the main file corresponding to the data manager
        fx_filename = os.path.join(filename, f'table.f{seqnr}')
        f = EndianAwareFileHandle(open(fx_filename, 'rb'), '>' if self.big_endian else '<', filename)

        # Open indirect array file if needed (sometimes arrays are stored
        # in these files).
        if os.path.exists(fx_filename + 'i'):
            fi = EndianAwareFileHandle(open(fx_filename + 'i', 'rb'), '>' if self.big_endian else '<', filename)
        else:
            fi = None

        # Start off by reading the bucket
        f.seek(512 + self.number_of_buckets * self.bucket_size + 4)
        index = ISMIndex.read(f)

        rows_in_bucket = np.diff(index.last_row.elements)
        rows_in_bucket = {key: value for (key, value) in zip(index.bucket_number.elements, rows_in_bucket)}

        n_rows = index.last_row.elements[-1]

        if n_rows > 0:

            data = []

            for bucket_id in index.bucket_number.elements:

                # Now move to the data location in the bucket
                f.seek(512 + bucket_id * self.bucket_size)

                # Read in length of data
                length = read_int32(f)

                # Read indices next to find out how many 'change' values there are
                f.seek(512 + bucket_id * self.bucket_size + length)

                for i in range(colindex_in_dm + 1):

                    n_changes = read_int32(f)

                    # Read in the indices
                    indices = np.frombuffer(f.read(n_changes * 4), dtype=f.endian + 'i4')

                    # Read in the offsets
                    offsets = np.frombuffer(f.read(n_changes * 4), dtype=f.endian + 'i4')

                # Now go back and read data
                f.seek(516 + bucket_id * self.bucket_size)

                values = []
                for off in offsets:
                    f.seek(516 + bucket_id * self.bucket_size + off)
                    if coldesc.is_direct or 'Scalar' in coldesc.stype:
                        values.append(read_as_numpy_array(f, coldesc.value_type, 1, length_modifier=-4))
                        subshape = []
                    else:
                        offset = read_int64(f)
                        fi.seek(offset)
                        ndim = read_int32(fi)
                        read_int32(fi)
                        subshape = []
                        for idim in range(ndim):
                            subshape.append(read_int32(fi))
                        size = int(prod(subshape))
                        values.append(read_as_numpy_array(fi, coldesc.value_type, size, shape=subshape[::-1]))
                if subshape:
                    values = np.vstack(values)
                else:
                    values = np.hstack(values)

                # Now expand into full size array

                # https://github.com/dask/dask/issues/4389
                repeats = np.diff(np.hstack([indices, rows_in_bucket[bucket_id]]))
                data.append(np.repeat(values, repeats, axis=0))

            if data[0].ndim > 1:
                return np.vstack(data)
            else:
                return np.hstack(data)

        else:

            return np.array([], dtype=TO_DTYPE[coldesc.value_type])



from math import prod

import numpy as np

from .table import CASATable
from .data_managers import StandardStMan, TiledCellStMan

__all__ = ['getdminfo', 'getdesc']


def getdminfo(filename, endian='>'):
    """
    Return the same output as CASA's getdminfo() function, namely a dictionary
    with metadata about the .image file, parsed from the ``table.f0`` file.
    """

    table = CASATable.read(filename, endian=endian)

    colset = table.column_set
    keys = list(colset.data_managers.keys())
    dm = colset.data_managers[keys[0]]

    dminfo = {}

    if isinstance(dm, StandardStMan):

        dminfo['COLUMNS'] = np.array(sorted(col.name for col in colset.columns), 'U')
        dminfo['NAME'] = dm.name
        dminfo['SEQNR'] = 0
        dminfo['TYPE'] = 'StandardStMan'

    dminfo['SPEC'] = {}

    if isinstance(dm, StandardStMan):

        dminfo['SPEC']['BUCKETSIZE'] = dm.bucket_size
        dminfo['SPEC']['IndexLength'] = dm.index_length
        dminfo['SPEC']['MaxCacheSize'] = dm.persistent_cache  # NOTE: not sure if correct
        dminfo['SPEC']['PERSCACHESIZE'] = dm.persistent_cache

    elif isinstance(dm, TiledCellStMan):

        dminfo['SPEC']['DEFAULTTILESHAPE'] = dm.default_tile_shape
        dminfo['SEQNR'] = dm.seqnr
        dminfo['SPEC']['SEQNR'] = dm.seqnr

        dminfo['COLUMNS'] = np.array([dm.column_name], dtype='<U16')
        dminfo['NAME'] = dm.column_name

        dminfo['SPEC']['MAXIMUMCACHESIZE'] = dm.max_cache_size
        dminfo['SPEC']['MaxCacheSize'] = dm.max_cache_size

        bucket = dminfo['SPEC']['HYPERCUBES'] = {}
        bucket = dminfo['SPEC']['HYPERCUBES']['*1'] = {}

        bucket['CubeShape'] = bucket['CellShape'] = dm.cube_shapes[0]
        bucket['TileShape'] = dm.tile_shapes[0]
        bucket['ID'] = {}
        bucket['BucketSize'] = int(dm.total_cube_size /
                                   prod(np.ceil(bucket['CubeShape']
                                              / bucket['TileShape'])))

        dminfo['TYPE'] = 'TiledCellStMan'

    return {'*1': dminfo}


def getdesc(filename, endian='>'):
    """
    Return the same output as CASA's getdesc() function, namely a dictionary
    with metadata about the .image file, parsed from the ``table.dat`` file.
    """

    table = CASATable.read(filename, endian=endian)

    coldesc = table.desc.column_description

    dmkey0 = list(table.column_set.data_managers.keys())[0]

    desc = {}
    for column in coldesc:
        desc[column.name] = {'comment': column.comment,
                             'dataManagerGroup': table.column_set.data_managers[dmkey0].name,
                             'dataManagerType': column.data_manager_type,
                             'keywords': column.keywords.as_dict(),
                             'maxlen': column.maxlen,
                             'option': column.option,
                             'valueType': column.value_type}
    desc['_keywords_'] = table.desc.keywords.as_dict()
    desc['_private_keywords_'] = table.desc.private_keywords.as_dict()
    desc['_define_hypercolumn_'] = {}

    return desc



# For convenience, we import the astropy Table class here which
# allows one to do:
#     >>> from casa_formats_io import Table
# instead of:
#     >>> import casa_formats_io
#     >>> from astropy.table import Table
from astropy.table import Table  # noqa

from .casa_dask import *  # noqa
from .casa_low_level_io import *  # noqa
from .casa_wcs import *  # noqa
from .version import version as __version__
from . import table_reader



# The helpers here make it possible to use dask arrays as columns in astropy
# tables.

# https://github.com/astropy/astropy/pull/12219

from astropy.utils.data_info import ParentDtypeInfo

import dask.array as da

__all__ = ['dask_to_mixin']


class DaskInfo(ParentDtypeInfo):
    @staticmethod
    def default_format(val):
        return f'{val.compute()}'


class DaskColumn(da.Array):

    info = DaskInfo()

    def copy(self):
        # Array hard-codes the resulting copied array as Array, so need to
        # overload this since Table tries to copy the array.
        return DaskColumn(self.dask, self.name, self.chunks, meta=self)

    def __getitem__(self, item):
        arr = super().__getitem__(item)
        return DaskColumn(arr.dask, arr.name, arr.chunks, meta=arr)



def dask_to_mixin(arr):
    return DaskColumn(arr.dask, arr.name, arr.chunks, meta=arr)



import struct
import warnings
from io import BytesIO

import numpy as np

TYPES = ['bool', 'char', 'uchar', 'short', 'ushort', 'int', 'uint', 'float',
         'double', 'complex', 'dcomplex', 'string', 'table', 'arraybool',
         'arraychar', 'arrayuchar', 'arrayshort', 'arrayushort', 'arrayint',
         'arrayuint', 'arrayfloat', 'arraydouble', 'arraycomplex',
         'arraydcomplex', 'arraystr', 'record', 'other']


class BaseCasaObject:
    def __repr__(self):
        from pprint import pformat
        return f'{self.__class__.__name__}' + pformat(self.__dict__)


class Block(BaseCasaObject):

    @classmethod
    def read(cls, f, func):
        self = cls()
        self.nr = read_int32(f)
        self.name = read_string(f)
        self.version = read_int32(f)
        self.size = read_int32(f)
        self.elements = [func(f) for i in range(self.size)]
        return self


class EndianAwareFileHandle:

    def __init__(self, file_handle, endian, original_filename):
        self.file_handle = file_handle
        self.endian = endian
        self.original_filename = original_filename

    def read(self, n=None):
        return self.file_handle.read(n)

    def tell(self):
        return self.file_handle.tell()

    def seek(self, n):
        return self.file_handle.seek(n)


def with_nbytes_prefix(func):
    def wrapper(*args):
        if hasattr(args[0], 'tell'):
            self = None
            f = args[0]
            args = args[1:]
        else:
            self = args[0]
            f = args[1]
            args = args[2:]
        # start = f.tell()
        nbytes = int(read_int32(f))
        if nbytes == 0:
            return
        bytes = f.read(nbytes - 4)
        b = EndianAwareFileHandle(BytesIO(bytes), f.endian,
                                  f.original_filename)
        if self:
            result = func(self, b, *args)
        else:
            result = func(b, *args)
        # end = f.tell()
        # if end - start != nbytes:
        #     raise IOError('Function {0} read {1} bytes instead of {2}'
        #                   .format(func, end - start, nbytes))
        return result
    return wrapper


def check_type_and_version(f, name, versions):

    # HACK: sometimes the endian flag is not set correctly on f, and we need to
    # figure out why. In the mean time, we can tell the actual endianness from
    # the next byte, because we expect the next four bytes to be the length of
    # the name string, and this won't be ridiculously long.

    start = f.tell()
    next = f.read(1)
    if next == b'\x00':
        actual_endian = '>'
    else:
        actual_endian = '<'
    f.seek(start)

    if actual_endian != f.endian:
        warnings.warn(f'Endianness of {name} did not match endianness of file'
                      'handle, correcting')
        f.endian = actual_endian

    if np.isscalar(versions):
        versions = [versions]
    stype, sversion = read_type(f)
    if stype != name or sversion not in versions:
        raise NotImplementedError('Support for {0} version {1} not implemented'
                                  .format(stype, sversion))

    return sversion


def read_bool(f):
    return f.read(1) == b'\x01'


def read_int16(f):
    return np.int16(struct.unpack(f.endian + 'h', f.read(2))[0])


def read_int32(f):
    return np.int32(struct.unpack(f.endian + 'i', f.read(4))[0])


def bytes_to_int32(bytes, endian):
    return np.int32(struct.unpack(endian + 'i', bytes)[0])


def read_int64(f):
    return np.int64(struct.unpack(f.endian + 'q', f.read(8))[0])


def read_float32(f):
    return np.float32(struct.unpack(f.endian + 'f', f.read(4))[0])


def read_float64(f):
    return np.float64(struct.unpack(f.endian + 'd', f.read(8))[0])


def read_complex64(f):
    return np.complex64(read_float32(f) + 1j * read_float32(f))


def read_complex128(f):
    return np.complex128(read_float64(f) + 1j * read_float64(f))


def read_string(f, length_modifier=0):
    value = read_int32(f) + length_modifier
    return f.read(int(value)).replace(b'\x00', b'').decode('ascii')


@with_nbytes_prefix
def read_iposition(f):
    check_type_and_version(f, 'IPosition', 1)
    nelem = read_int32(f)
    return np.array([read_int32(f) for i in range(nelem)], dtype=int)


ARRAY_ITEM_READERS = {
    'float': ('float', read_float32, np.float32),
    'double': ('double', read_float64, np.float64),
    'dcomplex': ('void', read_complex128, np.complex128),
    'string': ('String', read_string, 'U'),
    'int': ('Int', read_int32, int),
    'uint': ('uInt', read_int32, int)
}


TO_DTYPE = {}
TO_DTYPE['dcomplex'] = 'c16'
TO_DTYPE['complex'] = 'c8'
TO_DTYPE['double'] = 'f8'
TO_DTYPE['float'] = 'f4'
TO_DTYPE['int'] = 'i4'
TO_DTYPE['uint'] = 'u4'
TO_DTYPE['short'] = 'i2'
TO_DTYPE['string'] = 'U'
TO_DTYPE['bool'] = 'bool'
TO_DTYPE['record'] = 'O'

TO_TYPEREPR = {}
TO_TYPEREPR['dcomplex'] = 'void'
TO_TYPEREPR['double'] = 'double'
TO_TYPEREPR['float'] = 'float'
TO_TYPEREPR['int'] = 'Int'
TO_TYPEREPR['uint'] = 'uInt'
TO_TYPEREPR['string'] = 'String'


def read_as_numpy_array(f, value_type, nelem, shape=None, length_modifier=0):
    """
    Read the next 'nelem' values as a Numpy array
    """
    if value_type == 'string':
        array = np.array([read_string(f, length_modifier=length_modifier)
                          for i in range(nelem)])
    elif value_type == 'bool':
        length = int(np.ceil(nelem / 8)) * 8
        array = np.unpackbits(np.frombuffer(f.read(length), dtype='uint8'),
                              bitorder='little').astype(bool)[:nelem]
    elif value_type in TO_DTYPE:
        dtype = np.dtype(f.endian + TO_DTYPE[value_type])
        array = np.frombuffer(f.read(int(nelem * dtype.itemsize)), dtype=dtype)
    else:
        raise NotImplementedError(f"Can't read in data of type {value_type}")
    if shape is not None:
        array = array.reshape(shape)
    return array


@with_nbytes_prefix
def read_array(f, arraytype):

    typerepr = TO_TYPEREPR[arraytype]

    check_type_and_version(f, f'Array<{typerepr}>', 3)

    ndim = read_int32(f)
    shape = [read_int32(f) for i in range(ndim)]
    size = read_int32(f)

    return read_as_numpy_array(f, arraytype, size, shape=shape)


def read_type(f):
    tp = read_string(f)
    version = read_int32(f)
    return tp, version


@with_nbytes_prefix
def read_mapping(f, key_reader, value_reader):
    check_type_and_version(f, 'SimpleOrderedMap', 1)
    pos = f.tell()
    f.seek(pos)
    read_int32(f)  # ignored
    nr = read_int32(f)
    read_int32(f)  # ignored
    m = {}
    for i in range(nr):
        key = key_reader(f)
        value = value_reader(f)
        m[key] = value
    return m



import os
import pytest

DATA = os.path.join(os.path.dirname(__file__), '..', 'casa_low_level_io',
                    'tests', 'data')


def test_simple_direct_factory():
    pytest.importorskip('glue')
    # This import is deliberately here to not automatically register
    # the plugin for the whole test file - we want to make sure
    # in test_simple_load_data that the plugin registration works.
    from casa_formats_io.glue_factory import read_casa_table
    data = read_casa_table(os.path.join(DATA, 'simple.ms'))
    assert len(data) == 2
    assert data[0].shape == (10, 2)
    assert data[1].shape == (10, 4)


def test_simple_load_data():
    pytest.importorskip('glue')
    from glue.main import load_plugins
    from glue.core.data_factories import load_data
    load_plugins()
    data = load_data(os.path.join(DATA, 'simple.ms'))
    assert len(data) == 2
    assert data[0].shape == (10, 2)
    assert data[1].shape == (10, 4)



from .core import (check_type_and_version, BaseCasaObject, with_nbytes_prefix,
                   read_string, read_int32, read_iposition, TYPES)


class Record(BaseCasaObject):

    @classmethod
    @with_nbytes_prefix
    def read(cls, f):
        self = cls()
        check_type_and_version(f, 'Record', 1)
        self.desc = RecordDesc.read(f)
        read_int32(f)  # Not sure what the following value is
        return self


class RecordDesc(BaseCasaObject):

    @classmethod
    @with_nbytes_prefix
    def read(cls, f):

        self = cls()

        check_type_and_version(f, 'RecordDesc', 2)

        self.nrec = read_int32(f)

        self.names = []
        self.types = []
        self.values = []

        for i in range(self.nrec):
            self.names.append(read_string(f))
            self.types.append(TYPES[read_int32(f)])
            # Here we don't actually load in the data for may of the types - hence
            # why we don't do anything with the values we read in.
            if self.types[-1] in ('bool', 'int', 'uint', 'float', 'double',
                                  'complex', 'dcomplex', 'string'):
                self.values.append(read_string(f))  # noqa
            elif self.types[-1] == 'table':
                self.values.append(f.read(8))
            elif self.types[-1].startswith('array'):
                self.values.append(read_iposition(f))
                self.values.append(f.read(4))
            elif self.types[-1] == 'record':
                self.values.append(RecordDesc.read(f))
                self.values.append(read_int32(f))
            else:
                raise NotImplementedError("Support for type {0} in RecordDesc "
                                          "not implemented".format(self.types[-1]))

        return self



import os
import warnings
from math import prod
from io import BytesIO

import numpy as np

from ..core import (check_type_and_version, BaseCasaObject, with_nbytes_prefix,
                    read_string, read_int32, read_int64, read_as_numpy_array,
                    Block, EndianAwareFileHandle, TO_DTYPE, read_mapping,
                    bytes_to_int32)

__all__ = ['StandardStMan']


class SSMIndex(BaseCasaObject):

    @classmethod
    @with_nbytes_prefix
    def read(cls, f):
        self = cls()
        check_type_and_version(f, 'SSMIndex', 1)
        self.n_used = read_int32(f)
        self.rows_per_bucket = read_int32(f)
        self.n_columns = read_int32(f)
        self.free_space = read_mapping(f, read_int32, read_int32)

        self.last_row = Block.read(f, read_int32)
        self.bucket_number = Block.read(f, read_int32)

        return self


class StandardStMan(BaseCasaObject):

    @classmethod
    @with_nbytes_prefix
    def read(cls, f):
        self = cls()
        check_type_and_version(f, 'SSM', 2)
        self.name = read_string(f)
        self.column_offset = Block.read(f, read_int32)
        self.column_index_map = Block.read(f, read_int32)
        self._fileobj = f
        return self

    @with_nbytes_prefix
    def read_header(self, f):

        # SSMBase::readHeader()
        # https://github.com/casacore/casacore/blob/d6da19830fa470bdd8434fd855abe79037fda78c/tables/DataMan/SSMBase.cc#L415

        version = check_type_and_version(f, 'StandardStMan', (2, 3))

        if version >= 3:
            self.big_endian = f.read(1) == b'\x01'  # noqa
        else:
            self.big_endian = True

        self.bucket_size = read_int32(f)
        self.number_of_buckets = read_int32(f)
        self.persistent_cache = read_int32(f)
        self.number_of_free_buckets = read_int32(f)
        self.first_free_bucket = read_int32(f)
        self.number_of_bucket_for_index = read_int32(f)
        self.first_index_bucket_number = read_int32(f)
        self.idx_bucket_offset = read_int32(f)
        self.last_string_bucket = read_int32(f)
        self.index_length = read_int32(f)
        self.number_indices = read_int32(f)

    def read_column(self, filename, seqnr, column, coldesc, colindex_in_dm):

        # Open the main file corresponding to the data manager
        fx_filename = os.path.join(filename, f'table.f{seqnr}')
        f = EndianAwareFileHandle(open(fx_filename, 'rb'), '>' if self.big_endian else '<', filename)

        # Open indirect array file if needed (sometimes arrays are stored
        # in these files).
        if os.path.exists(fx_filename + 'i'):
            fi = EndianAwareFileHandle(open(fx_filename + 'i', 'rb'), '>' if self.big_endian else '<', filename)
        else:
            fi = None

        # Variable length strings are stored in their own buckets which we cache
        # as needed.
        variable_string_buckets = {}
        next_variable_string_buckets = {}

        def _ensure_variable_string_bucket_loaded(f, vs_bucket_id):
            if vs_bucket_id in variable_string_buckets:
                return next_variable_string_buckets[vs_bucket_id]
            pos = f.tell()
            f.seek(512 + self.bucket_size * vs_bucket_id + 12)
            # For some reason, the next bucket index is stored in big endian
            next_vs_bucket_id = bytes_to_int32(f.read(4), '>')
            variable_string_buckets[vs_bucket_id] = f.read(self.bucket_size - 16)
            next_variable_string_buckets[vs_bucket_id] = next_vs_bucket_id
            f.seek(pos)
            return next_vs_bucket_id

        # To read in the SSMIndex we need to pre-load the index buckets. In
        # cases where the index is split over multiple buckets, the first four
        # bytes of the bucket indicate the index of the next index bucket to
        # read. This corresponds to the logic in SSMBase.readIndexBuckets:
        # https://github.com/casacore/casacore/blob/master/tables/DataMan/SSMBase.cc#L454
        # We pre-load the index buckets into a single BytesIO to make it look
        # contiguous.

        index_bytes = b''
        next_index_bucket = self.first_index_bucket_number
        remaining_index_length = self.index_length

        for bucket_id in range(self.number_of_bucket_for_index):

            bucket_start = 512 + next_index_bucket * self.bucket_size

            f.seek(bucket_start)

            # For some reason, the next bucket index is stored in big endian
            next_index_bucket = bytes_to_int32(f.read(4), '>')

            if self.idx_bucket_offset > 0:
                f.seek(bucket_start + self.idx_bucket_offset)
                index_bytes += f.read(remaining_index_length)
            elif remaining_index_length < self.bucket_size:
                f.seek(bucket_start + 8)
                index_bytes += f.read(remaining_index_length)
            else:
                f.seek(bucket_start + 8)
                index_bytes += f.read(self.bucket_size - 8)

        index_bytes = EndianAwareFileHandle(BytesIO(index_bytes[4:]), f.endian, f.original_filename)

        index = SSMIndex.read(index_bytes)

        if index.bucket_number.elements == []:  # empty table
            if coldesc.value_type in TO_DTYPE:
                return np.array([], dtype=TO_DTYPE[coldesc.value_type])
            else:
                return None

        shape = column.data.shape
        nelements = int(prod(shape))

        data = []

        rows_in_bucket = np.diff(np.hstack([0, np.array(index.last_row.elements) + 1]))
        rows_in_bucket = {key: max(0, value) for (key, value) in zip(index.bucket_number.elements, rows_in_bucket)}

        for bucket_id in index.bucket_number.elements:

            # Find the starting position of the column in the bucket
            f.seek(512 + self.bucket_size * (bucket_id + self.column_index_map.elements[colindex_in_dm]) + self.column_offset.elements[colindex_in_dm])

            if coldesc.value_type == 'string':
                if coldesc.maxlen == 0:
                    subdata = []
                    for irow in range(rows_in_bucket[bucket_id]):
                        bytes = f.read(8)
                        length = read_int32(f)
                        if length <= 8:
                            subdata.append(bytes[:length])
                        else:
                            vs_bucket_id = bytes_to_int32(bytes[:4], f.endian)
                            offset = bytes_to_int32(bytes[4:], f.endian)
                            next_vs_bucket_id = _ensure_variable_string_bucket_loaded(f, vs_bucket_id)
                            bytes = variable_string_buckets[vs_bucket_id][offset:offset + length]
                            if len(bytes) < length:
                                _ensure_variable_string_bucket_loaded(f, next_vs_bucket_id)
                                bytes += variable_string_buckets[next_vs_bucket_id][:length - len(bytes)]
                            if coldesc.ndim != 0:
                                if coldesc.is_fixed_shape:
                                    n = prod(coldesc.shape)
                                    pos = 0
                                else:
                                    n = bytes_to_int32(bytes[4:8], '>')
                                    pos = 12
                                strings = []
                                for i in range(n):
                                    length = bytes_to_int32(bytes[pos:pos + 4], '>')
                                    strings.append(bytes[pos + 4: pos + 4 + length])
                                    pos += 4 + length
                                if coldesc.is_fixed_shape:
                                    strings = np.reshape(strings, coldesc.shape)
                                bytes = strings
                            subdata.append(bytes)
                    try:
                        data.append(np.array(subdata))
                    except ValueError:
                        data.append(np.array(subdata, dtype=object))
                else:
                    data.append(np.frombuffer(f.read(coldesc.maxlen * rows_in_bucket[bucket_id]), dtype=f'S{coldesc.maxlen}'))
            elif coldesc.value_type == 'record':
                # TODO: determine how to handle this properly
                warnings.warn(f'Skipping column {coldesc.name} with type record')
                data = None
            else:
                if coldesc.is_direct or 'Scalar' in coldesc.stype:
                    data.append(read_as_numpy_array(f, coldesc.value_type, rows_in_bucket[bucket_id] * nelements, shape=(-1,) + shape[::-1]))
                else:
                    values = []
                    for irow in range(rows_in_bucket[bucket_id]):
                        offset = read_int64(f)
                        fi.seek(offset)
                        ndim = read_int32(fi)
                        subshape = []
                        for idim in range(ndim):
                            subshape.append(read_int32(fi))
                        size = int(prod(subshape))
                        values.append(read_as_numpy_array(fi, coldesc.value_type, size, shape=subshape[::-1]))
                    try:
                        data.append(np.array(values))
                    except ValueError:
                        data.append(np.array(values, dtype=object))
        if data:
            if data[0].ndim > 1:
                return np.vstack(data)
            else:
                return np.hstack(data)
        else:
            return None



def _peek(f, length):
    # Internal function used for debugging - show the next N bytes (and the
    # previous 8)
    pos = f.tell()
    f.seek(pos - 8)
    print(repr(f.read(8)), '  ', repr(f.read(length)))
    f.seek(pos)


def _peek_no_prefix(f, length):
    # Internal function used for debugging - show the next N bytes
    pos = f.tell()
    print(repr(f.read(length)))
    f.seek(pos)



#!/usr/bin/env python

import os
import sys

import numpy

from setuptools import setup
from setuptools.extension import Extension

setup(use_scm_version={'write_to': os.path.join('casa_formats_io', 'version.py')},
      ext_modules=[Extension("casa_formats_io._casa_chunking",
                             [os.path.join('casa_formats_io', '_casa_chunking.c')],
                             py_limited_api=True,
                             include_dirs=[numpy.get_include()])])



import numpy as np

from ..core import (check_type_and_version, BaseCasaObject, with_nbytes_prefix,
                    read_string, read_int32, TYPES, read_as_numpy_array)

__all__ = ['StManAipsIO']


class StManAipsIO(BaseCasaObject):

    @classmethod
    def read(cls, f):
        self = cls()
        self.name = 'StManAipsIO'
        return self

    @with_nbytes_prefix
    def read_header(self, f):
        version = check_type_and_version(f, 'StManAipsIO', 2)
        if version > 1:
            self.name = read_string(f)
        self.seqnr = read_int32(f)
        self.unique_number = read_int32(f)
        self.nrow = read_int32(f)
        self.ncol = read_int32(f)
        self.value_types = [TYPES[read_int32(f)] for icol in range(self.ncol)]
        self.columns = [StManColumnAipsIO.read(f, self.value_types[icol]) for icol in range(self.ncol)]

    def read_column(self, filename, seqnr, column, coldesc, colindex_in_dm):
        return self.columns[colindex_in_dm].values


class StManColumnAipsIO(BaseCasaObject):

    @classmethod
    def read(cls, f, value_type):
        self = cls()
        read_int32(f)
        check_type_and_version(f, 'StManColumnAipsIO', 2)
        self.nr = read_int32(f)
        irow = 0
        self.values = []
        while irow < self.nr:
            nr = read_int32(f)
            nr = read_int32(f)
            self.values.append(read_as_numpy_array(f, value_type, nr))
            irow += nr
        self.values = np.hstack(self.values)
        return self



import os
import warnings
from collections import OrderedDict

import numpy as np

from dask import array as da

from astropy.table import Table as AstropyTable

from .core import (check_type_and_version, BaseCasaObject, with_nbytes_prefix,
                   read_string, read_int32, read_bool, read_iposition,
                   EndianAwareFileHandle, read_type, TYPES, read_array,
                   read_float32, read_float64, read_complex64, read_complex128)

from .data_managers import (StandardStMan, IncrementalStMan, TiledCellStMan,
                            TiledShapeStMan, TiledColumnStMan, StManAipsIO,
                            VariableShapeArrayList)

from .record import RecordDesc

from .dask_mixin import dask_to_mixin


class ArrayColumnData(BaseCasaObject):

    @classmethod
    def read(cls, f):

        self = cls()

        self.version = read_int32(f)
        self.seqnr = read_int32(f)

        has_shape = read_bool(f)

        if has_shape:
            self.shape = tuple(read_iposition(f).tolist())
        else:
            self.shape = ()

        return self


class ScalarColumnData(BaseCasaObject):

    @classmethod
    def read(cls, f):

        self = cls()

        self.version = read_int32(f)
        self.seqnr = read_int32(f)
        self.shape = ()

        return self


class ColumnDesc(BaseCasaObject):

    @classmethod
    def read(cls, f):

        self = cls()

        unknown = read_int32(f)  # noqa

        stype, sversion = read_type(f)

        if not stype.startswith(('ScalarColumnDesc', 'ScalarRecordColumnDesc', 'ArrayColumnDesc')) or sversion != 1:
            raise NotImplementedError('Support for {0} version {1} not implemented'.format(stype, sversion))

        # https://github.com/casacore/casacore/blob/dbf28794ef446bbf4e6150653dbe404379a3c429/tables/Tables/BaseColDesc.cc#L285

        self.stype = stype
        self.name = read_string(f)
        self.comment = read_string(f)
        self.data_manager_type = read_string(f).replace('Shape', 'Cell')
        self.data_manager_group = read_string(f)
        self.value_type = TYPES[read_int32(f)]

        self.option = read_int32(f)
        self.is_direct = self.option & 1 == 1
        self.is_undefined = self.option & 2 == 2
        self.is_fixed_shape = self.option & 4 == 4

        self.ndim = read_int32(f)
        if self.ndim != 0:
            self.shape = read_iposition(f)  # noqa
        self.maxlen = read_int32(f)
        self.keywords = TableRecord.read(f)

        read_int32(f)
        if 'ArrayColumnDesc' in stype:
            sw = f.read(1)  # noqa
        else:
            if self.value_type in ('ushort', 'short'):
                self._default = f.read(2)
            elif self.value_type in ('uint', 'int', 'float'):
                self._default = f.read(4)
            elif self.value_type in ('double', 'complex'):
                self._default = f.read(8)
            elif self.value_type in ('dcomplex'):
                self._default = f.read(16)
            elif self.value_type == 'bool':
                self._default = f.read(1)
            elif self.value_type == 'string':
                self._default = read_string(f)
            elif self.value_type == 'record':
                self._default = f.read(8)
            else:
                raise NotImplementedError(f"Can't read default value for {self.value_type}")

        return self


class PlainColumn(BaseCasaObject):

    @classmethod
    def read(cls, f, ndim):

        self = cls()

        version = read_int32(f)

        if version < 2:
            raise NotImplementedError('Support for PlainColumn version {0} not implemented'.format(version))

        self.name = read_string(f)

        if ndim != 0:
            self.data = ArrayColumnData.read(f)
        else:
            self.data = ScalarColumnData.read(f)

        return self


class ColumnSet(BaseCasaObject):

    @classmethod
    def read(cls, f, desc):

        self = cls()

        version = read_int32(f)  # can be negative
        # See full logic in ColumnSet.getFile
        version = -version

        if version != 2:
            raise NotImplementedError('Support for ColumnSet version {0} not implemented'.format(version))

        self.nrow = read_int32(f)
        self.nrman = read_int32(f)
        self.nr = read_int32(f)

        # Construct data managers

        data_manager_cls = OrderedDict()

        for i in range(self.nr):

            name = read_string(f)
            seqnr = read_int32(f)

            if name == 'StandardStMan':
                dm_cls = StandardStMan
            elif name == 'IncrementalStMan':
                dm_cls = IncrementalStMan
            elif name == 'TiledCellStMan':
                dm_cls = TiledCellStMan
            elif name == 'TiledShapeStMan':
                dm_cls = TiledShapeStMan
            elif name == 'TiledColumnStMan':
                dm_cls = TiledColumnStMan
            elif name == 'StManAipsIO':
                dm_cls = StManAipsIO
            else:
                raise NotImplementedError('Data manager {0} not supported'.format(name))

            data_manager_cls[seqnr] = dm_cls

        self.columns = [PlainColumn.read(f, ndim=desc.column_description[index].ndim) for index in range(desc.ncol)]

        # Prepare data managers

        f.read(8)  # includes a length in bytes and bebebebe, need to check how this behaves when multiple DMs are present

        self.data_managers = OrderedDict()

        for seqnr in data_manager_cls:
            self.data_managers[seqnr] = data_manager_cls[seqnr].read(f)
            f.read(8)

        return self


class TableRecord(BaseCasaObject):

    @classmethod
    @with_nbytes_prefix
    def read(cls, f):

        self = cls()

        check_type_and_version(f, 'TableRecord', 1)

        desc = RecordDesc.read(f)

        unknown = read_int32(f)  # noqa

        self.values = {}

        for name, rectype in zip(desc.names, desc.types):
            if rectype == 'bool':
                self.values[name] = read_bool(f)
            elif rectype == 'int':
                self.values[name] = int(read_int32(f))
            elif rectype == 'uint':
                self.values[name] = int(read_int32(f))
            elif rectype == 'float':
                self.values[name] = float(read_float32(f))
            elif rectype == 'double':
                self.values[name] = float(read_float64(f))
            elif rectype == 'complex':
                self.values[name] = complex(read_complex64(f))
            elif rectype == 'dcomplex':
                self.values[name] = complex(read_complex128(f))
            elif rectype == 'string':
                self.values[name] = read_string(f)
            elif rectype == 'table':
                self.values[name] = 'Table: ' + os.path.abspath(os.path.join(f.original_filename, read_string(f)))
            elif rectype == 'arrayint':
                self.values[name] = read_array(f, 'int').astype('<i4')
            elif rectype == 'arrayuint':
                self.values[name] = read_array(f, 'uint').astype('<u4')
            elif rectype == 'arrayfloat':
                self.values[name] = read_array(f, 'float').astype('<f4')
            elif rectype == 'arraydouble':
                self.values[name] = read_array(f, 'double').astype('<f8')
            elif rectype == 'arraycomplex':
                self.values[name] = read_array(f, 'complex').astype('<c8')
            elif rectype == 'arraydcomplex':
                self.values[name] = read_array(f, 'dcomplex').astype('<c16')
            elif rectype == 'arraystr':
                self.values[name] = read_array(f, 'string')
            elif rectype == 'record':
                self.values[name] = TableRecord.read(f).values
            else:
                raise NotImplementedError("Support for type {0} in TableRecord not implemented".format(rectype))

        return self

    def as_dict(self):
        return self.values


def ensure_mixin_columns(columns):
    new_columns = {}
    for colname, column in columns.items():
        if isinstance(column, da.Array):
            new_columns[colname] = dask_to_mixin(column)
        else:
            new_columns[colname] = column
    return new_columns


class CASATable(BaseCasaObject):

    @classmethod
    def read(cls, filename, endian='<'):
        """
        Read a CASA table - note that this is currently **experimental** and
        should not be used for production code.
        """

        with open(os.path.join(filename, 'table.dat'), 'rb') as f_orig:

            f = EndianAwareFileHandle(f_orig, '>', filename)

            magic = f.read(4)
            if magic != b'\xbe\xbe\xbe\xbe':
                raise ValueError('Incorrect magic code: {0}'.format(magic))

            table = cls.read_fileobj(f)

        for dm_index, dm in table.column_set.data_managers.items():

            fx_filename = os.path.join(filename, f'table.f{dm_index}')

            if os.path.exists(fx_filename):

                with open(fx_filename, 'rb') as f_orig:

                    if isinstance(dm, (TiledCellStMan, TiledShapeStMan, StManAipsIO)):
                        endian = '>'

                    f = EndianAwareFileHandle(f_orig, endian, filename)

                    magic = f.read(4)
                    if magic != b'\xbe\xbe\xbe\xbe':
                        raise ValueError('Incorrect magic code: {0}'.format(magic))

                    dm.read_header(f)

        table._filename = filename

        return table

    def has_data_desc_id(self):
        coldesc = self.desc.column_description
        for colindex in range(len(coldesc)):
            colname = coldesc[colindex].name
            if colname == 'DATA_DESC_ID':
                return True
        return False

    def as_astropy_table(self, data_desc_id=None, include_columns=None):
        """
        Parameters
        ----------
        data_desc_id : 'all', None, or int
            'all' will get all of the ddids as a list of tables
            An integer will get the specified ddid number.
            If there is only one ddid and `data_desc_id` is None, it will return that ddid.
            Otherwise, an exception will be raised.

        Returns
        -------
        table or list of tables
        """

        # We now loop over columns and read the relevant data from each bucket.

        # TODO: refactor this to make it so that columns are read on request
        # instead of all in one go. The best way to do this might be to make a
        # table where each column is a dask array.

        coldesc = self.desc.column_description

        table_columns = OrderedDict()

        for colindex in range(len(coldesc)):

            colname = coldesc[colindex].name

            if include_columns is not None and colname not in include_columns:
                continue

            # Find the data manager to use for the column as well as the
            # 'sequence number' - this is the value in e.g. table.f<seqnr>
            seqnr = self.column_set.columns[colindex].data.seqnr
            dm = self.column_set.data_managers[seqnr]

            # Each data manager might only handle one or a few of the columns.
            # It may internally have a list of the columns it deals with, so
            # we need to figure out what the column index is in that specific
            # data manager
            colindex_in_dm = 0
            for column in self.column_set.columns[:colindex]:
                if column.data.seqnr == seqnr:
                    colindex_in_dm += 1

            if hasattr(dm, 'read_column'):
                coldata = dm.read_column(self._filename, seqnr, self.column_set.columns[colindex], coldesc[colindex], colindex_in_dm)
                if coldata is not None:
                    table_columns[colname] = coldata
            else:
                warnings.warn(f'Skipping column {colname} with data manager {dm.__class__.__name__}')

        # In some files (e.g. ms files) some columns can have variable shape. For now
        # we only deal with this if DATA_DESC_ID is defined as a column in the table
        # as it allows us to know how to extract sub-tables with constant shapes.

        if 'DATA_DESC_ID' in table_columns:

            # If DATA_DESC_ID is present but include_columns is used, it might
            # actually be possible to represent the data as a single table.
            for data in table_columns.values():
                if isinstance(data, VariableShapeArrayList):
                    break
            else:
                return AstropyTable(data=ensure_mixin_columns(table_columns))

            data_desc_ids = sorted(np.unique(table_columns['DATA_DESC_ID']))

            if data_desc_id is None and len(data_desc_ids) == 1:
                data_desc_id = data_desc_ids[0]
            elif data_desc_id == 'all':
                tables = [self._read_data_descid(table_columns, ddid)
                          for ddid in data_desc_ids]
                return tables

            try:
                return self._read_data_descid(table_columns, data_desc_id)
            except Exception as ex:
                raise ValueError("There are multiple DATA_DESC_ID values present "
                                 "in the table, select the one you need with "
                                 "data_desc_id=<value>. Valid options are "
                                 + "/".join([str(x) for x in data_desc_ids])
                                 + f"  Parent exception was: {ex}")


        else:

            for data in table_columns.values():
                if isinstance(data, VariableShapeArrayList):
                    raise NotImplementedError("Can't handle tables with variable shaped columns where DATA_DESC_ID does not exist")

            return AstropyTable(data=ensure_mixin_columns(table_columns))

    def _read_data_descid(self, table_columns, data_desc_id):

        keep = table_columns['DATA_DESC_ID'] == data_desc_id

        columns_sub = OrderedDict()
        for colname, data in table_columns.items():
            if isinstance(data, VariableShapeArrayList):
                if len(data) == 0:
                    continue
                for row_index, array in data:
                    # Workaround for https://github.com/dask/dask/issues/8387
                    if len(row_index) < 32:
                        row_index = row_index.compute()
                    matches = table_columns['DATA_DESC_ID'][row_index] == data_desc_id
                    if np.any(matches):
                        columns_sub[colname] = array[matches]
                        break
                else:
                    columns_sub[colname] = []
            else:
                columns_sub[colname] = data[keep]

        table = AstropyTable(data=ensure_mixin_columns(columns_sub))
        table.meta['DATA_DESC_ID'] = data_desc_id

        return table

    @classmethod
    @with_nbytes_prefix
    def read_fileobj(cls, f):

        self = cls()

        check_type_and_version(f, 'Table', 2)

        self.nrow = read_int32(f)
        self.fmt = read_int32(f)  # noqa
        self.name = read_string(f)  # noqa

        # big_endian = fmt == 0  # noqa

        self.desc = TableDesc.read(f, self.nrow)

        self.column_set = ColumnSet.read(f, desc=self.desc)

        return self


class TableDesc(BaseCasaObject):

    @classmethod
    @with_nbytes_prefix
    def read(cls, f, nrow):

        self = cls()

        check_type_and_version(f, 'TableDesc', 2)

        unknown1 = read_string(f)  # noqa
        unknown2 = read_string(f)  # noqa
        unknown3 = read_string(f)  # noqa

        self.keywords = TableRecord.read(f)
        self.private_keywords = TableRecord.read(f)

        self.ncol = read_int32(f)

        self.column_description = []

        for icol in range(self.ncol):
            self.column_description.append(ColumnDesc.read(f))

        return self
